/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_OBLD_DEBUG			FALSE

#include "grass/grass.pmh"

using namespace fe;

void BladeOp::initialize(void)
{
	//* NOTE inherits attributes from SurfaceBindOp

	catalogRemove("Spider");

	catalog<String>("icon")="FE_beta";

	//* Page Config

	catalog<String>("Threads","page")="Config";

	catalog<String>("AutoThread","page")="Config";

	catalog<String>("StayAlive","page")="Config";

	catalog<String>("Tangent Attribute")="tangent";
	catalog<String>("Tangent Attribute","page")="Config";
	catalog<String>("Tangent Attribute","hint")=
			"Name of vector primitive attribute on input curves"
			" describing the direction along which rotations can occur.";

	catalog<String>("Facing Attribute")="facing";
	catalog<String>("Facing Attribute","page")="Config";
	catalog<String>("Facing Attribute","hint")=
			"Name of vector point attribute added to the input curves"
			" describing the facing side."
			"  These may get rotated in the output.";

	catalog<String>("Force Attribute")="force";
	catalog<String>("Force Attribute","page")="Config";
	catalog<String>("Force Attribute","hint")=
			"Name of vector point attribute on input curves"
			" describing an external force.";

	catalog<String>("Velocity Attribute")="v";
	catalog<String>("Velocity Attribute","page")="Config";
	catalog<String>("Velocity Attribute","hint")=
			"Name of vector point attribute on input curves"
			" describing relative air movement."
			"  Velocity values should produce a better wind response than"
			" an external force since the solver can account for facing area.";

	//* Page Physics

	catalog<Real>("Tension")=1e4;
	catalog<Real>("Tension","high")=1e5;
	catalog<String>("Tension","page")="Physics";
	catalog<String>("Tension","hint")="Angular springiness.";

	catalog<Real>("Drag")=1e2;
	catalog<Real>("Drag","high")=1e3;
	catalog<String>("Drag","page")="Physics";
	catalog<String>("Drag","hint")="Angular sluggishness.";

	catalog<Real>("Mass")=1e0;
	catalog<Real>("Mass","high")=1e1;
	catalog<String>("Mass","page")="Physics";
	catalog<String>("Mass","hint")="Physical presence per vertex point.";

	catalog<Real>("Viscosity")=1e0;
	catalog<Real>("Viscosity","high")=1e1;
	catalog<String>("Viscosity","page")="Physics";
	catalog<String>("Viscosity","hint")="Effectiveness of air velocity.";

	catalog<Real>("Restoration")=0.0;
	catalog<Real>("Restoration","high")=1e5;
	catalog<String>("Restoration","page")="Physics";

	catalog<Real>("Time Step")=1.0/60.0;
	catalog<Real>("Time Step","high")=1e-1;
	catalog<String>("Time Step","page")="Physics";

	catalog<Real>("Wind")=0;
	catalog<String>("Wind","label")="Test Wind";
	catalog<Real>("Wind","high")=1e3;
	catalog<String>("Wind","page")="Physics";

	catalog<Real>("Turbulence")=0;
	catalog<String>("Turbulence","label")="Test Turbulence";
	catalog<Real>("Turbulence","high")=1e3;
	catalog<String>("Turbulence","page")="Physics";

	catalog<Real>("Bend")=1e-2;
	catalog<String>("Bend","label")="Added Bend";
	catalog<Real>("Bend","high")=1e0;
	catalog<String>("Bend","page")="Physics";
	catalog<String>("Bend","hint")=
			"Flatness curvature.  This is uniformly added to all rest angles.";

#if FE_CODEGEN>FE_DEBUG
	catalog<bool>("Wind","hidden")=true;
	catalog<bool>("Turbulence","hidden")=true;
	catalog<bool>("Bend","hidden")=true;
#endif

	//* Page Driver

	catalog<bool>("Driver Bound")=false;
	catalog<String>("Driver Bound","page")="Driver";
	catalog<String>("Driver Bound","hint")=
			"This option requires the reference Driver Surface."
			"  An optional animated Deformed Surface should match"
			" the topology of the Driver Surface.";

	catalogMoveToEnd("DriverGroup");
	catalog<String>("DriverGroup","enabler")="Driver Bound";
	catalog<String>("DriverGroup","page")="Driver";

	catalogMoveToEnd("Bindings");
	catalog<I32>("Bindings","max")=1;
	catalog<String>("Bindings","enabler")="Driver Bound";
	catalog<String>("Bindings","page")="Driver";
	catalog<bool>("Bindings","hidden")=true;

	catalogMoveToEnd("Bind Primitives");
	catalog<String>("Bind Primitives","enabler")="Driver Bound";
	catalog<String>("Bind Primitives","page")="Driver";
	catalog<bool>("Bind Primitives","hidden")=true;

	catalogMoveToEnd("Bind Fragments");
	catalog<String>("Bind Fragments","enabler")="Driver Bound";
	catalog<String>("Bind Fragments","page")="Driver";
	catalog<bool>("Bind Fragments","hidden")=true;

	catalogMoveToEnd("BindPrefix");
	catalog<String>("BindPrefix","enabler")="Driver Bound";
	catalog<String>("BindPrefix","page")="Driver";

	catalogMoveToEnd("Bind Center");
	catalog<String>("Bind Center","enabler")="Driver Bound";
	catalog<String>("Bind Center","page")="Driver";
	catalog<bool>("Bind Center","hidden")=true;

	catalogMoveToEnd("FragmentAttr");
	catalog<String>("FragmentAttr","enabler")="Driver Bound";
	catalog<String>("FragmentAttr","page")="Driver";
	catalog<bool>("FragmentAttr","hidden")=true;

	//* Page Collision

	catalog<String>("Collision Method")="Ray Cast Halvsies";
	catalog<String>("Collision Method","choice:0")="BoundingSphere";
	catalog<String>("Collision Method","label:0")=
			"Bounding Sphere (vertex collisions only)";
	catalog<String>("Collision Method","choice:1")="NearestPoint";
	catalog<String>("Collision Method","label:1")="Nearest Point";
	catalog<String>("Collision Method","choice:2")="RayCastOneStep";
	catalog<String>("Collision Method","label:2")="Ray Cast - One Step";
	catalog<String>("Collision Method","choice:3")="RayCastHalvsies";
	catalog<String>("Collision Method","label:3")="Ray Cast - Halvsies";
	catalog<String>("Collision Method","page")="Collision";

	catalog<Real>("Collision Gap")=0.0;
	catalog<String>("Collision Gap","page")="Collision";
	catalog<String>("Collision Gap","hint")=
			"Imposed extra distance from the collider.";

	catalog<I32>("Refinement")=0;
	catalog<I32>("Refinement","high")=4;
	catalog<I32>("Refinement","max")=8;
	catalog<String>("Refinement","page")="Collision";
	catalog<String>("Refinement","hint")=
			"Maximum level of collider subdivision."
			"  Refinement may be adaptive and non-uniform.";

	catalog<bool>("Grabbing")=false;
	catalog<String>("Grabbing","page")="Collision";

	catalog<Real>("Grab Bias")=0.0;
	catalog<Real>("Grab Bias","min")= -1.0;
	catalog<Real>("Grab Bias","max")=1.0;
	catalog<String>("Grab Bias","enabler")="Grabbing";
	catalog<String>("Grab Bias","page")="Collision";
	catalog<String>("Grab Bias","hint")=
			"Tunes how easily the blade tips stick to the collider.";

	catalog<bool>("Debug Surface")=false;
	catalog<String>("Debug Surface","page")="Collision";
	catalog<String>("Debug Surface","hint")=
			"Draw lines over the output showing the"
			" collider surface refinement."
			"  This feedback may only be visible is this node is"
			" \"selected\" and/or \"displayed\".";

	catalog<bool>("Debug Collisions")=false;
	catalog<String>("Debug Collisions","page")="Collision";
	catalog<String>("Debug Collisions","hint")=
			"Add lines to the output detailing the collisions."
			"Draw lines over the output detailing the collisions."
			"  This feedback may only be visible is this node is"
			" \"selected\" and/or \"displayed\".";

	//* Page Buckling

	catalog<bool>("Buckling")=false;
	catalog<String>("Buckling","page")="Buckling";

	catalog<Real>("Break Angle")=1.0;
	catalog<Real>("Break Angle","min")=0.0;
	catalog<Real>("Break Angle","max")=2.0;
	catalog<String>("Break Angle","enabler")="Buckling";
	catalog<String>("Break Angle","page")="Buckling";
	catalog<String>("Break Angle","hint")=
			"Threshold above which buckling will occur.";

	catalog<Real>("Elastic Angle")=0.1;
	catalog<Real>("Elastic Angle","min")=0.0;
	catalog<Real>("Elastic Angle","max")=2.0;
	catalog<String>("Elastic Angle","enabler")="Buckling";
	catalog<String>("Elastic Angle","page")="Buckling";
	catalog<String>("Elastic Angle","hint")=
			"Subtracted from Break Angle when reseting dynamic rest angle.";

	catalog<Real>("Pop Angle")=0.8;
	catalog<Real>("Pop Angle","min")=0.0;
	catalog<Real>("Pop Angle","max")=2.0;
	catalog<String>("Pop Angle","enabler")="Buckling";
	catalog<String>("Pop Angle","page")="Buckling";
	catalog<String>("Pop Angle","hint")=
			"Threshold below which buckling is fully recovered.";

	catalog<Real>("Recovery Angle")=0.01;
	catalog<Real>("Recovery Angle","min")=0.0;
	catalog<Real>("Recovery Angle","max")=0.1;
	catalog<String>("Recovery Angle","enabler")="Buckling";
	catalog<String>("Recovery Angle","page")="Buckling";
	catalog<String>("Recovery Angle","hint")=
			"Incremental movement of dynamic rest angle towards"
			" original rest angle.";

	catalog<String>("Input Surface","label")="Input Curves";

	catalog< sp<Component> >("Collider");
	catalog<bool>("Collider","optional")=true;

	catalogMoveToEnd("Driver Surface");
	catalog<bool>("Driver Surface","optional")=true;

	catalog< sp<Component> >("Deformed Surface");
	catalog<bool>("Deformed Surface","optional")=true;

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","label")="Output Curves";
	catalog<String>("Output Surface","IO")="output";
	catalog<String>("Output Surface","copy")="Input Surface";
	catalog<bool>("Output Surface","recycle")=true;

//	catalogDump();

	m_spDebugGroup=new DrawMode();
	m_spDebugGroup->setGroup("debug");
}

void BladeOp::handle(Record& a_rSignal)
{
	const I32 frame=currentFrame(a_rSignal);
	const BWORD replaced=catalog<bool>("Output Surface","replaced");

#if FE_OBLD_DEBUG
	feLog("BladeOp::handle frame %d replaced %d\n",frame,replaced);
#endif

	//* output curves are copy of input curves
	if(!accessOutput(m_spOutputAccessible,a_rSignal)) return;

	sp<SurfaceI> spCollider;
	access(spCollider,"Collider",e_warning);

	sp<SurfaceI> spDriver;
	sp<SurfaceI> spDeformed;
	if(catalog<bool>("Driver Bound"))
	{
		if(!access(spDriver,"Driver Surface")) return;

		spDriver->setRefinement(0);
		spDriver->center();

		if(access(spDeformed,"Deformed Surface",e_quiet))
		{
			spDeformed->setRefinement(0);
			spDeformed->center();
		}
		else
		{
			spDeformed=spDriver;
		}
	}
	else
	{
		spDriver=NULL;
		spDeformed=NULL;
	}

	if(spCollider.isValid())
	{
		const I32 refinement=catalog<I32>("Refinement");
		spCollider->setRefinement(refinement);
	}

	const I32 threads=catalog<I32>("Threads");
	const BWORD debugSurface=catalog<bool>("Debug Surface");
	const BWORD debugCollisions=catalog<bool>("Debug Collisions");
	const BWORD recycle=catalog<bool>("Output Surface","recycle");
	const BWORD wantDebug=
			((debugSurface || debugCollisions) && threads==1);

	sp<DrawI> spDrawDebug;
	if(wantDebug && !recycle && !accessGuide(spDrawDebug,a_rSignal)) return;

	if(recycle==wantDebug)
	{
		catalog<bool>("Output Surface","recycle")=(!wantDebug);
	}

	m_blade.setDrawDebug(debugCollisions? spDrawDebug: sp<DrawI>(NULL));

	m_blade.setCollider(spCollider);
	m_blade.setDriver(spDriver);
	m_blade.setDeformed(spDeformed);

	String tangentName=catalog<String>("Tangent Attribute");
	String facingName=catalog<String>("Facing Attribute");
	String forceName=catalog<String>("Force Attribute");
	String velocityName=catalog<String>("Velocity Attribute");

	m_blade.setTension(catalog<Real>("Tension"));
	m_blade.setRestoration(catalog<Real>("Restoration"));
	m_blade.setDrag(catalog<Real>("Drag"));
	m_blade.setBend(catalog<Real>("Bend"));
	m_blade.setGrabbing(catalog<bool>("Grabbing"));
	m_blade.setGrabBias(catalog<Real>("Grab Bias"));
	m_blade.setCollisionGap(catalog<Real>("Collision Gap"));
	m_blade.setMass(catalog<Real>("Mass"));
	m_blade.setWind(catalog<Real>("Wind"));
	m_blade.setTurbulence(catalog<Real>("Turbulence"));
	m_blade.setBuckling(catalog<bool>("Buckling"));
	m_blade.setBreakAngle(catalog<Real>("Break Angle"));
	m_blade.setElasticAngle(catalog<Real>("Elastic Angle"));
	m_blade.setPopAngle(catalog<Real>("Pop Angle"));
	m_blade.setRecoveryAngle(catalog<Real>("Recovery Angle"));

	const String collisionString=catalog<String>("Collision Method");
	Blade::CollisionMethod collisionMethod=Blade::e_boundingSphere;
	if(collisionString=="NearestPoint")
	{
		collisionMethod=Blade::e_nearestPoint;
	}
	else if(collisionString=="RayCastOneStep")
	{
		collisionMethod=Blade::e_rayCastOneStep;
	}
	else if(collisionString=="RayCastHalvsies")
	{
		collisionMethod=Blade::e_rayCastHalvsies;
	}
	m_blade.setCollisionMethod(collisionMethod);

	const Real timeStep=catalog<Real>("Time Step");
	m_blade.setTimeStep(timeStep);

	static Real t=0.0;	// HACK
	t+=timeStep;
	m_blade.setTime(t);

	m_blade.populate();

	sp<SurfaceAccessorI> spPrimitiveTangent;
	access(spPrimitiveTangent,m_spOutputAccessible,
			e_primitive,tangentName,e_quiet,e_refuseMissing);

	sp<SurfaceAccessorI> spPointFacing;
	access(spPointFacing,m_spOutputAccessible,
			e_point,facingName,e_quiet,e_refuseMissing);

	sp<SurfaceAccessorI> spPointForce;
	access(spPointForce,m_spOutputAccessible,
			e_point,forceName,e_quiet,e_refuseMissing);

	sp<SurfaceAccessorI> spPointVelocity;
	access(spPointVelocity,m_spOutputAccessible,
			e_point,velocityName,e_quiet,e_refuseMissing);

	sp<SurfaceAccessorI> spInputPosition;
	if(!access(spInputPosition,"Input Surface",
			e_point,e_position)) return;

	sp<SurfaceAccessorI> spPrimitiveVertices;
	if(!access(spPrimitiveVertices,m_spOutputAccessible,
			e_primitive,e_vertices)) return;

	const U32 pointCount=spInputPosition->count();
	const U32 primitiveCount=spPrimitiveVertices->count();
	const U32 subCount=primitiveCount? pointCount/primitiveCount: 0;

	std::vector<SpatialVector>& rTangent=m_blade.tangent();
	std::vector<SpatialVector>& rFacing=m_blade.facing();
	std::vector<SpatialVector>& rForceIn=m_blade.forceIn();
	std::vector<SpatialVector>& rVelocityIn=m_blade.velocityIn();
	std::vector<SpatialVector>& rLocationRef=m_blade.locationRef();
	std::vector<Real>& rVelocity=m_blade.velocity();
	std::vector<I32>& rBindFace=m_blade.bindFace();
	std::vector<SpatialBary>& rBindBary=m_blade.bindBary();

	//* inferred reset
	const BWORD reuseCache=(frame==m_frame+1 || (frame==m_frame && !replaced));

#if FE_OBLD_DEBUG
	if(reuseCache)
	{
		if(frame==m_frame+1)
		{
			feLog("BladeOp::handle temporal step\n");
		}
		else
		{
			feLog("BladeOp::handle temporal repeat\n");
		}
	}
#endif

	m_frame=frame;

	if(!reuseCache)
	{
#if FE_OBLD_DEBUG
		feLog("BladeOp::handle reset cache\n");
#endif
		rLocationRef.resize(pointCount);
		for(U32 m=0;m<pointCount;m++)
		{
			rLocationRef[m]=spInputPosition->spatialVector(m,0);
		}

		const U32 pointCountV=subCount? pointCount*(subCount-1)/subCount: 0;
		rVelocity.resize(pointCountV);
		for(U32 m=0;m<pointCountV;m++)
		{
			rVelocity[m]=0.0;
		}

		rFacing.resize(pointCount);
		if(spPointFacing.isValid() &&
				spPointFacing->count() &&
				spPointFacing->subCount(0))
		{
			for(U32 m=0;m<pointCount;m++)
			{
				rFacing[m]=spPointFacing->spatialVector(m,0);
			}
		}
		else
		{
//			feLog("BladeOp::handle facing attribute \"%s\" missing\n",
//					facingName.c_str());
			catalog<String>("warning")="facing attribute \"" + facingName +
					"\" is missing; ";

			for(U32 m=0;m<pointCount;m++)
			{
				set(rFacing[m]);
			}
		}

		const U32 pointCountT=subCount? pointCount/subCount: 0;
		rTangent.resize(pointCountT);
		rBindFace.resize(pointCountT);
		rBindBary.resize(pointCountT);
		if(spPrimitiveTangent.isValid() &&
				spPrimitiveTangent->count() && spPrimitiveTangent->subCount(0))
		{
			for(U32 m=0;m<pointCountT;m++)
			{
				rTangent[m]=spPrimitiveTangent->spatialVector(m,0);
			}
		}
		else
		{
//			feLog("BladeOp::handle tangent attribute \"%s\" missing\n",
//					tangentName.c_str());
			catalog<String>("warning")+="tangent attribute \"" + tangentName +
					"\" is missing; ";

			for(U32 m=0;m<pointCountT;m++)
			{
				const Real angle=sin(0.7*m)+sin(0.3*m);

				set(rTangent[m],cos(angle),0.0,sin(angle));
			}
		}
		std::vector< sp<SurfaceAccessorI> > outputFaces;
		std::vector< sp<SurfaceAccessorI> > outputBarys;
		const I32 bindings=accessBindings(m_spOutputAccessible,
				outputFaces,outputBarys,FALSE);
		if(bindings<1)
		{
//			feLog("BladeOp::handle no binding data\n");
			catalog<String>("warning")+="binding data is missing; ";
		}

		if(bindings>0 && outputFaces[0].isValid() &&
				outputBarys[0].isValid())
		{
			for(U32 m=0;m<pointCountT;m++)
			{
				rBindFace[m]=outputFaces[0]->integer(m,0);
				rBindBary[m]=outputBarys[0]->spatialVector(m,0);
			}
		}
		else
		{
//			feLog("BladeOp::handle bind attributes not both present\n");
			if(bindings)
			{
				catalog<String>("warning")+="binding data is incomplete; ";
			}

			for(U32 m=0;m<pointCountT;m++)
			{
				rBindFace[m]= -1;
				set(rBindBary[m]);
			}
		}
		for(U32 m=0;m<pointCountT;m++)
		{
			m_blade.storeAsRest(m);
		}
	}

	rForceIn.resize(pointCount);
	if(spPointForce.isValid() &&
			spPointForce->count() &&
			spPointForce->subCount(0))
	{
		for(U32 m=0;m<pointCount;m++)
		{
			rForceIn[m]=spPointForce->spatialVector(m,0);
		}
	}
	else
	{
//		feLog("BladeOp::handle force attribute \"%s\" missing\n",
//				forceName.c_str());
		catalog<String>("message")+="force attribute \"" + forceName +
				"\" is missing; ";

		for(U32 m=0;m<pointCount;m++)
		{
			set(rForceIn[m]);
		}
	}

	rVelocityIn.resize(pointCount);
	if(spPointVelocity.isValid() &&
			spPointVelocity->count() &&
			spPointVelocity->subCount(0))
	{
		const Real viscosity=catalog<Real>("Viscosity");
		for(U32 m=0;m<pointCount;m++)
		{
			rVelocityIn[m]=viscosity*spPointVelocity->spatialVector(m,0);
		}
	}
	else
	{
//		feLog("BladeOp::handle velocity attribute \"%s\" missing\n",
//				velocityName.c_str());
		catalog<String>("message")+="velocity attribute \"" + velocityName +
				"\" is missing; ";

		for(U32 m=0;m<pointCount;m++)
		{
			set(rVelocityIn[m]);
		}
	}

	if(spDrawDebug.isValid())
	{
		spDrawDebug->pushDrawMode(m_spDebugGroup);
	}

	setFullRange(0,primitiveCount-1);
	OperatorThreaded::handle(a_rSignal);

	if(spDrawDebug.isValid())
	{
		if(debugSurface)
		{
			sp<DrawableI> spDrawable=spCollider;
			if(spDrawable.isValid())
			{
				spDrawable->draw(spDrawDebug,NULL);
			}
		}

		spDrawDebug->popDrawMode();
	}

#if FE_OBLD_DEBUG
	feLog("BladeOp::handle done\n");
#endif
}

void BladeOp::run(I32 a_id,Range a_range)
{
	const String facingName=catalog<String>("Facing Attribute");

	sp<SurfaceAccessorI> spPointFacing;
	if(!access(spPointFacing,m_spOutputAccessible,e_point,facingName)) return;

	sp<SurfaceAccessorI> spOutputPosition;
	if(!access(spOutputPosition,m_spOutputAccessible,
			e_point,e_position)) return;

	sp<SurfaceAccessorI> spPrimitiveVertices;
	if(!access(spPrimitiveVertices,m_spOutputAccessible,
			e_primitive,e_vertices)) return;

	const U32 pointCount=spOutputPosition->count();
	const U32 primitiveCount=spPrimitiveVertices->count();
	const U32 subCount=primitiveCount? pointCount/primitiveCount: 0;

	const I32 start=a_range.m_start;
	const I32 end=a_range.m_end;

	std::vector<SpatialVector>& rLocationDef=m_blade.locationDef();
	std::vector<SpatialVector>& rFacing=m_blade.facing();

#if FE_OBLD_DEBUG
	feLog("BladeOp::run running thread %d [%d,%d]\n",a_id,start,end);
#endif

	for(U32 index=start;index<=end;index++)
	{
		m_blade.step(index);

		const U32 startSub=index*subCount;
		const U32 endSub=startSub+subCount;
		for(U32 indexSub=startSub;indexSub<endSub;indexSub++)
		{
			spOutputPosition->set(indexSub,0,rLocationDef[indexSub]);
		}
		if(spPointFacing.isValid() &&
				spPointFacing->count() &&
				spPointFacing->subCount(0))
		{
			for(U32 indexSub=startSub;indexSub<endSub;indexSub++)
			{
				spPointFacing->set(indexSub,0,rFacing[indexSub]);
			}
		}
	}

#if FE_OBLD_DEBUG
	feLog("BladeOp::run done running thread %d\n",a_id);
#endif
}
